//////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2005
//   Nicholas Kidd  <kidd@cs.wisc.edu>
//   Thomas Reps    <reps@cs.wisc.edu>
//   David Melski   <melski@grammatech.com>
//   Akash Lal      <akash@cs.wisc.edu>
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   1. Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//
//   2. Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//
//   3. Neither the name of the University of Wisconsin, Madison,
//      nor GrammaTech, Inc., nor the names of the copyright holders
//      and contributors may be used to endorse or promote
//      products derived from this software without specific prior
//      written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//////////////////////////////////////////////////////////////////////////////
/*
 * $Id: ref_ptr.h,v 1.31 2005/06/09 01:48:49 kidd Exp $
 */
#ifndef REF_PTR_H_
#define REF_PTR_H_
#include <stdio.h>
#include <cassert>
#include <limits.h>
#include <iostream>
#include <list>

/* A reference counting pointer class
 *
 * This class is *NOT* thread safe.
 *
 * The templated class must have a member variable named count
 * that can be accessed from ref_ptr and modified by ref_ptr.
 * the count variable should have operator++(),operator--(), and
 * operator==( int ) defined.  As a note, this class was designed
 * with count being an unsigned integer, although there may be some
 * advantages to making it be of type RefCounter, as implemented in
 * this file. (E.g., it handles the warnings in the next paragraph.)
 *
 * Count should be initialized to 0 for proper reference
 * couting to work.  If it is desirable for the pointer/object
 * to never be deleted, initialize count >= 1.  WARNING: don't forget
 * to initialize count to 0 in the copy constructor!  (And don't
 * forget to have a copy constructor.) The assignment operator should
 * probably not change the value either.
 *
 * ref_ptr provides a templated function operator<< that allows 
 * the use of std::ostream << with ref_ptr<T> objects.  To take
 * advantage of this, the templated class must provide a method
 * with signature:
 *
 * 		std::ostream& print( std::ostream& )
 *
 */

class RefCounter {
    friend std::ostream& operator<<( std::ostream& o, const RefCounter& rc )
    {
        return (o << rc.cnt);
    }
 public:
    /* Constructor: by default, when "born," no one references us. */
    RefCounter(unsigned _cnt = 0) : cnt(_cnt) { }

    /* Copy Constructor: we are being born, and by default, will have
       no referrers --- even if the object we are copying has
       referrers. */
    RefCounter(const RefCounter &that) : cnt(0) { }

    RefCounter& operator++() { ++cnt; return *this; }
    RefCounter& operator--() { --cnt; return *this; }

    /* The assignment operator: we are copying the 'value' of
       "that", but we will still have our own set of referrers when
       done. (We are not copying "that's" referrers.) */
    RefCounter & operator=(const RefCounter &that) { return *this; }

    bool operator==(unsigned i) const { return (unsigned)i == cnt; }
 private:
    unsigned cnt;
};

inline bool operator==(unsigned i, const RefCounter &rc) { return rc == i; }

template< typename T > class ref_ptr {

    public:
        typedef unsigned int count_t;

        ref_ptr( T *t = 0 ) {
            acquire( t );
        }

        ref_ptr( const ref_ptr& rp ) {
            acquire( rp.ptr );
        }

        ~ref_ptr() {
            release();
        }

		ref_ptr& operator=( T* t ) {
		  if( ptr != t ) {
              T* old_ptr = ptr;
              acquire(t);
              release(old_ptr);
          }
		  return *this;
		}

        ref_ptr& operator=( const ref_ptr& rp ) {
            return *this = rp.ptr;
        }

        bool operator==(const ref_ptr& that) const {
            return ptr == that.ptr;
        }

        T * get_ptr() const {
            return ptr;
        }

        T * operator->() const {
            return ptr;
        }

        T& operator*()   const {
#ifdef DBGREFPTR
            if(0 == ptr)
            {
                std::cerr << "ref_ptr::operator*: Invalid ptr.\n";
            }
#endif
            /* for NDEBUG. Make sure it crashes
               here rather than have some
               undefined behaviour. */
            assert(0 != ptr);
            return *ptr;
        }

        std::ostream& print( std::ostream& out )
        {
            if( ptr )
                ptr->print( out );
            else
                out << "NULL";
            return out;
        }

        bool is_empty() const { return (0 == ptr); }
        bool is_valid() const { return !(is_empty());}

    private:
        T *ptr;

        void acquire( T *t ) {
            ptr = t;
            if( t ) {
                ++t->count;
#ifdef DBGREFPTR
                std::cout << "Acquired " << t << " with count = "
                    << t->count << std::endl;
#endif
            }
        }

#ifdef GRAMMATECH_TRACEBACK
        // WARNING: This use of deathRow probably makes this code "non-reentrant".
        static std::list<T*> l_deathRow; 

        static void release(T* old_ptr)
        {

            if( old_ptr ) {
                --old_ptr->count;
                if( old_ptr->count == 0 ) {
                    if( l_deathRow.empty() ) {
                        // We want to delete the pointed to object, but we need
                        // to be careful about creating a deep recursion if the
                        // pointed to object is a recursive data structure that
                        // has ref_ptr<T> members.  So, we begin a deathRow,
                        // to handle the recursion.
                        l_deathRow.push_front(old_ptr);
                        while( ! l_deathRow.empty() ) {
                            // Check, just to make sure it hasn't been
                            // "reclaimed" somehow.
                            if( l_deathRow.front()->count == 0 )
                                // This delete may cause deathRow to grow:
                                delete l_deathRow.front(); 
                            l_deathRow.pop_front();
                        }
                    } else {
                        // The deathRow is not empty, which means that we are
                        // in a recursive delete, so we simply add our child to
                        // the end, and count on the "parent" to handle it.
                        l_deathRow.push_back(old_ptr);
                    }
                }
            }
        }
#else
        static void release(T* old_ptr)
        {
            if( old_ptr ) {
                --old_ptr->count;
#ifdef DBGREFPTR
                std::cout << "Released " << *old_ptr << " with count = "
                    << old_ptr->count << std::endl;
#endif
                if( old_ptr->count == 0 ) {
#ifdef DBGREFPTR
                    std::cout << "Deleting ptr: " << *old_ptr << std::endl;
#endif
                    delete old_ptr;
                }
            }
        }
#endif

        void release() {
		  release(ptr);
		  ptr = 0;
		}
};

/*
template< typename T > std::ostream& operator<<(
		std::ostream& o,
		const ref_ptr< T >& r )
{
	if( 0 != r.get_ptr() )
		return r->print(o);
	else
		return (o << "NULL");
}
*/

template< typename T > std::ostream& operator<<(
		std::ostream& o,
        const ref_ptr< T > r )
{
	if( 0 != r.get_ptr() )
		return r->print(o);
	else
		return (o << "NULL");
}

#endif	// REF_PTR_H_
/* Yo, Emacs!
;;; Local Variables: ***
;;; tab-width: 4 ***
;;; End: ***
*/
/*
 * $Log: ref_ptr.h,v $
 * Revision 1.31  2005/06/09 01:48:49  kidd
 * Changed the CND_DBGS in ref_ptr::operator* to "ifdef DBGREFPTR". This removes
 * false dependence of ref_ptr on common.h. Fixed Dictionary::clear() so that the
 * memory associated with the "values" vector is reclaimed. This involved the use
 * of vector<T>::swap and a temporary. Many kudos to Ben for the tip.
 *
 * Revision 1.30  2005/05/04 15:52:06  melski
 * Just added some comments to RefCounter.
 *
 * Revision 1.29  2005/04/14 15:31:43  kidd
 *
 * Added:
 *     typedef unsigned int count_t
 *
 * User semiring implementations can define count to be of type
 * ref_ptr<T>::count_t. This allows for replacing unsigned int with something
 * else (like the RefCounter class defined in ref_ptr.h)
 *
 * Revision 1.28  2005/03/09 22:18:57  kidd
 * Added the file COPYRIGHT and inserted this file at the beginning of each h and
 * cpp file.  Changed default CFLAG in Makefile to be "-ggdb -DDBGWPDS". This
 * initializes the integer "b_*_debug" flags to 1; however, turning compile time
 * warnings on or off is still successful as templates are instantiated at
 * compile time.
 *
 */
